@@ -39,7 +39,9 @@
  *                    CMS,
  *                    CMSCondMark,
  *                    Serial,
- *                    Parallel}
+ *                    Parallel,
+ *                    Shenandoah,
+ *                    ShenandoahIU}
  */
 
 
@@ -100,6 +102,18 @@ public class TestVolatiles {
             procArgs[argcount - 3] = "-XX:+UseConcMarkSweepGC";
             procArgs[argcount - 2] = "-XX:+UseCondCardMark";
             break;
+        case "Shenandoah":
+            argcount = 9;
+            procArgs = new String[argcount];
+            procArgs[argcount - 2] = "-XX:+UseShenandoahGC";
+            break;
+        case "ShenandoahIU":
+            argcount = 11;
+            procArgs = new String[argcount];
+            procArgs[argcount - 4] = "-XX:+UnlockExperimentalVMOptions";
+            procArgs[argcount - 3] = "-XX:+UseShenandoahGC";
+            procArgs[argcount - 2] = "-XX:ShenandoahGCMode=iu";
+            break;
         default:
             throw new RuntimeException("unexpected test type " + testType);
         }
@@ -355,6 +369,17 @@ public class TestVolatiles {
                     "ret"
                 };
                 break;
+            case "Shenandoah":
+            case "ShenandoahIU":
+                 // Shenandoah generates normal object graphs for
+                 // volatile stores
+                matches = new String[] {
+                    "membar_release \\(elided\\)",
+                    useCompressedOops ? "stlrw?" : "stlr",
+                    "membar_volatile \\(elided\\)",
+                    "ret"
+                };
+                break;
             }
         } else {
             switch (testType) {
@@ -418,6 +443,20 @@ public class TestVolatiles {
                     "ret"
                 };
                 break;
+
+            case "Shenandoah":
+            case "ShenandoahIU":
+                 // Shenandoah generates normal object graphs for
+                 // volatile stores
+                matches = new String[] {
+                    "membar_release",
+                    "dmb ish",
+                    useCompressedOops ? "strw?" : "str",
+                    "membar_volatile",
+                    "dmb ish",
+                    "ret"
+                };
+                break;
             }
         }
 
@@ -520,6 +559,17 @@ public class TestVolatiles {
                     "ret"
                 };
                 break;
+            case "Shenandoah":
+            case "ShenandoahIU":
+                // For volatile CAS, Shenanodoah generates normal
+                // graphs with a shenandoah-specific cmpxchg
+                matches = new String[] {
+                    "membar_release \\(elided\\)",
+                    useCompressedOops ? "cmpxchgw?_acq_shenandoah" : "cmpxchg_acq_shenandoah",
+                    "membar_acquire \\(elided\\)",
+                    "ret"
+                };
+                break;
             }
         } else {
             switch (testType) {
@@ -583,6 +633,17 @@ public class TestVolatiles {
                     "ret"
                 };
                 break;
+            case "Shenandoah":
+            case "ShenandoahIU":
+                // For volatile CAS, Shenanodoah generates normal
+                // graphs with a shenandoah-specific cmpxchg
+                matches = new String[] {
+                    "membar_release",
+                    useCompressedOops ? "cmpxchgw?_shenandoah" : "cmpxchg_shenandoah",
+                    "membar_acquire",
+                    "ret"
+                };
+                break;
             }
         }
 
@@ -764,6 +825,19 @@ public class TestVolatiles {
                     "ret"
                 };
                 break;
+            case "Shenandoah":
+            case "ShenandoahIU":
+                // For volatile CAS, Shenanodoah generates normal
+                // graphs with a shenandoah-specific cmpxchg
+                matches = new String[] {
+                    "membar_release",
+                    "dmb ish",
+                    useCompressedOops ? "cmpxchgw?_shenandoah" : "cmpxchg_shenandoah",
+                    "membar_acquire",
+                    "dmb ish",
+                    "ret"
+                };
+                break;
             }
         }
 
